/*******************************************************************************
*                                 Sudoku                                       *
********************************************************************************
*            Copyright (C) 2006 Sander Jansen. All Rights Reserved             *
*                               ---                                            *
* This program is free software; you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation; either version 2 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    *
*******************************************************************************/
/*
$Id$
*/
#include <fx.h>
#include "sudoku.h"


Sudoku::Sudoku() : num_solved(0) {
  for (int i=0;i<81;i++) field[i]=0;
  }

Sudoku::Sudoku(const Sudoku & s ) {
  for (int i=0;i<81;i++)
    field[i]=s.field[i];
  num_solved=s.num_solved;
  }

Sudoku& Sudoku::operator=(const Sudoku& s) {
  for (int i=0;i<81;i++)
    field[i]=s.field[i];
  num_solved=s.num_solved;
  return *this;
  }

bool Sudoku::operator==(const Sudoku& s) const {
  for (int i=0;i<81;i++) {
    if (field[i]!=s.field[i])
      return false;
    }
  return true;
  }

bool Sudoku::operator!=(const Sudoku& s) const {
  for (int i=0;i<81;i++) {
    if (field[i]!=s.field[i])
      return true;
    }
  return false;
  }

void Sudoku::clear() {
  for (int i=0;i<81;i++) field[i]=0;
  num_solved=0;
  }

FXchar Sudoku::get(int r,int c) const {
  return field[(r*9)+c];
  }

FXchar Sudoku::get(int i) const {
  return field[i];
  }

void Sudoku::set(int r,int c,FXchar value) {
  set((r*9)+c,value);
  }

void Sudoku::set(int i,FXchar value) {
  if (value!=field[i]) {
    if (value>0 && field[i]==0)
      num_solved++;
    else if (value==0 && field[i]>0)
      num_solved--;

    field[i]=value;
    }
  }

bool Sudoku::solved() const{
  return (num_solved==81);
  }
  
bool Sudoku::empty() const{
  return (num_solved==0);  
  }


bool Sudoku::is_valid(int pos,int value){
  int row = pos / 9;
  int col = pos % 9;

  for (int i=(row*9);i<(row*9)+9;i++){
    if (field[i]==value) {
      return false;
      }
    }

  for (int i=0;i<9;i++){
    if (field[(i*9)+col]==value){
      return false;
      }
    }

  int sr = row / 3;
  int sc = col / 3;
  for (int i=0;i<3;i++){
    for (int j=0;j<3;j++){
      if (field[(sr*3*9)+(sc*3)+(i*9)+j]==value){
        return false;
        }
      }
    }
  return true;
  }

void Sudoku::save(FXStream & store) const{
  store.save(field,81);
  }

void Sudoku::load(FXStream & store) {
  store.load(field,81);
  num_solved=0;
  for (int i=0;i<81;i++){
    field[i]=FXCLAMP(0,field[i],9);
    if (field[i]>0)
      num_solved++;
    }
  }



