/*******************************************************************************
*                                 Sudoku                                       *
********************************************************************************
*            Copyright (C) 2006 Sander Jansen. All Rights Reserved             *
*                               ---                                            *
* This program is free software; you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation; either version 2 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    *
*******************************************************************************/
/*
$Id$
*/
#include <new>
#include <fx.h>
#include <FXArray.h>
#include "dlx.h"
#include "sudoku.h"
#include "sudokusolver.h"

SudokuSolver::SudokuSolver() : dlx(324,2916,81) {
  }

bool SudokuSolver::solve(Sudoku & sudoku){
  FXArray<dlxnode*> rows;
  dlxnode * row;
  int data[4];
  int sindex=0;

  for (int i=0;i<9;i++){
    for (int j=0;j<9;j++){
      for (int k=0;k<9;k++){
        // 1. Each cell in a row and column must contain exactly one digit: <r,c>.
        // 2. Each digit must occur in each row exactly once: <d,r>.
        // 3. Each digit must occur in each column exactly once: <d,c>.
        // 4. Each digit must occur in each box exactly once: <d,b>.
        data[0]=1+(i*9+j);                  // <r,c> = columns: 1 <> 81
        data[1]=82+(i*9+k);                 // <d,r> = columns: 82 <> 162
        data[2]=163+(j*9+k);                // <d,c> = columns: 163 <> 243
        data[3]=244+(((i/3)*3+(j/3))*9+k);  // <d,b> = columns: 244 <> 325
        row = initrow(data,4);
        if (sudoku.get(i,j)==k+1)
          rows.append(row);
        }
      }
    }

  for (int i=0;i<rows.no();i++){
    cover(rows[i]->column);
    removeRow(rows[i]);
    solution[sindex++]=rows[i];
    }

  if (findFirst(sindex)) {
    int r,c,v;
    for (int i=0;i<81;i++){
      r=c=v=-1;
      dlxnode * row = solution[i];
      do {
        if (row->column->name<=81) {
          r = (row->column->name-1) / 9;
          c = (row->column->name-1) % 9;
          }
        else if (row->column->name<=162){
          v = (row->column->name-1) % 9;
          }
        row=row->right;
        }
      while(row!=solution[i]);
      if (r==-1 || c==-1 || v==-1)
        return false;

      sudoku.set(r,c,v+1);
      }
    return true;
    }
  return false;
  }

