/*******************************************************************************
*                                 Sudoku                                       *
********************************************************************************
*            Copyright (C) 2006 Sander Jansen. All Rights Reserved             *
*                               ---                                            *
* This program is free software; you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation; either version 2 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    *
*******************************************************************************/
/*
$Id$
*/
#include <fx.h>
#include <FXPNGIcon.h>
#include "dlx.h"
#include "sudokusolver.h"
#include "sudokumove.h"
#include "annotationmove.h"
#include "sudokuwindow.h"
#include "aboutwindow.h"
#include "icons.h"
#include "version.h"

FXDEFMAP(SudokuWindow) SudokuWindowMap[]={
  FXMAPFUNC(SEL_UPDATE,SudokuWindow::ID_SOLVE,SudokuWindow::onUpdSolve),
  FXMAPFUNC(SEL_UPDATE,SudokuWindow::ID_SAVE,SudokuWindow::onUpdSave),

  FXMAPFUNCS(SEL_VERIFY,SudokuWindow::ID_ANNOTATION_0,SudokuWindow::ID_ANNOTATION_80,SudokuWindow::onCmdAnnotation),
  //  FXMAPFUNCS(SEL_CHANGED,SudokuWindow::ID_ANNOTATION_0,SudokuWindow::ID_ANNOTATION_80,SudokuWindow::onCmdAnnotation),

  FXMAPFUNCS(SEL_COMMAND,SudokuWindow::ID_FIELD_0,SudokuWindow::ID_FIELD_80,SudokuWindow::onCmdField),
  FXMAPFUNCS(SEL_VERIFY,SudokuWindow::ID_FIELD_0,SudokuWindow::ID_FIELD_80,SudokuWindow::onVerifyField),
  FXMAPFUNCS(SEL_CHANGED,SudokuWindow::ID_FIELD_0,SudokuWindow::ID_FIELD_80,SudokuWindow::onCmdField),
  FXMAPFUNC(SEL_COMMAND,SudokuWindow::ID_SOLVE,SudokuWindow::onCmdSolve),
  FXMAPFUNC(SEL_COMMAND,SudokuWindow::ID_LOAD,SudokuWindow::onCmdLoad),
  FXMAPFUNC(SEL_COMMAND,SudokuWindow::ID_SAVE,SudokuWindow::onCmdSave),

  FXMAPFUNC(SEL_COMMAND,SudokuWindow::ID_CLEAR,SudokuWindow::onCmdClear),
  FXMAPFUNC(SEL_COMMAND,SudokuWindow::ID_RESET,SudokuWindow::onCmdReset),

  FXMAPFUNC(SEL_COMMAND,SudokuWindow::ID_ABOUT,SudokuWindow::onCmdAbout),
  };

FXIMPLEMENT(SudokuWindow,FXMainWindow,SudokuWindowMap,ARRAYNUMBER(SudokuWindowMap));

SudokuWindow::SudokuWindow(FXApp * app) : FXMainWindow(app,"Sudoku",NULL,NULL,DECOR_TITLE|DECOR_MINIMIZE|DECOR_MAXIMIZE|DECOR_CLOSE|DECOR_BORDER|DECOR_MENU){
  FXFontDesc fontdesc;
  getApp()->getNormalFont()->getFontDesc(fontdesc);

  fontdesc.size=(FXushort)(fontdesc.size * 1.5);
  largefont=new FXFont(getApp(),fontdesc);

  fontdesc.size=(FXushort)(fontdesc.size / 2.0);
  smallfont=new FXFont(getApp(),fontdesc);


  icon_exit = new FXPNGIcon(getApp(),exit_png);
  icon_undo = new FXGIFIcon(getApp(),undo_gif);
  icon_redo = new FXGIFIcon(getApp(),redo_gif);
  icon_logo = new FXPNGIcon(getApp(),logo_png);
  icon_logo_small = new FXGIFIcon(getApp(),logo_small_gif);
  icon_logo_about = new FXPNGIcon(getApp(),logo_about_png);

  setIcon(icon_logo);
  setMiniIcon(icon_logo_small);


  FXMenuBar * menubar = new FXMenuBar(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X);

    /// File Menu
  filemenu = new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Game",NULL,filemenu);
  new FXMenuCommand(filemenu,"Open...",NULL,this,SudokuWindow::ID_LOAD);
  new FXMenuCommand(filemenu,"Save...",NULL,this,SudokuWindow::ID_SAVE);
  new FXMenuCommand(filemenu,"Solve",NULL,this,SudokuWindow::ID_SOLVE);
  new FXMenuSeparator(filemenu);
  new FXMenuCommand(filemenu,"Reset",NULL,this,SudokuWindow::ID_RESET);
  new FXMenuCommand(filemenu,"Clear",NULL,this,SudokuWindow::ID_CLEAR);
  new FXMenuSeparator(filemenu);
  new FXMenuCommand(filemenu,"&Quit\tCtrl-Q\tExit application",icon_exit,this,FXTopWindow::ID_CLOSE);

  editmenu = new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Edit",NULL,editmenu);
  new FXMenuCommand(editmenu,tr("Undo\tCtrl-Z\tUndo last change."),icon_undo,&undolist,FXUndoList::ID_UNDO);
  new FXMenuCommand(editmenu,tr("Redo\tCtrl-Y\tRedo last undo."),icon_redo,&undolist,FXUndoList::ID_REDO);

  helpmenu = new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Help",NULL,helpmenu,LAYOUT_RIGHT);
  new FXMenuCommand(helpmenu,"&About Sudoku",NULL,this,ID_ABOUT);



  FXVerticalFrame * vframe;
  FXMatrix * matrix;
  FXMatrix * mainmatrix = new FXMatrix(this,3,PACK_UNIFORM_WIDTH|LAYOUT_CENTER_X|LAYOUT_CENTER_Y|PACK_UNIFORM_HEIGHT|MATRIX_BY_COLUMNS|FRAME_SUNKEN,0,0,0,0,2,2,2,2,2,2);

  for (int h=0;h<3;h++){
    for (int i=0;i<3;i++){
      matrix = new FXMatrix(mainmatrix,3,LAYOUT_FILL_X|PACK_UNIFORM_WIDTH|PACK_UNIFORM_HEIGHT|PACK_UNIFORM_WIDTH|MATRIX_BY_COLUMNS,0,0,0,0,0,0,0,0,1,1);
      matrix->setBackColor(FXRGB(0,0,0));
      for (int j=0;j<3;j++){
        for (int k=0;k<3;k++){
          vframe = new FXVerticalFrame(matrix,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,0,0,0,0,0,0);
          pv[(h*3*9)+(i*3)+(j*9)+k] = new FXTextField(vframe,6,this,(h*3*9)+(i*3)+(j*9)+k+ID_ANNOTATION_0,LAYOUT_FILL_X|TEXTFIELD_LIMITED|TEXTFIELD_INTEGER);
          pv[(h*3*9)+(i*3)+(j*9)+k]->setFont(smallfont);
          field[(h*3*9)+(i*3)+(j*9)+k] = new FXTextField(vframe,1,this,(h*3*9)+(i*3)+(j*9)+k+ID_FIELD_0,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_CENTER_X|TEXTFIELD_LIMITED|TEXTFIELD_INTEGER,0,0,0,0,0,0,0,6);
          field[(h*3*9)+(i*3)+(j*9)+k]->setJustify(JUSTIFY_CENTER_X);
          field[(h*3*9)+(i*3)+(j*9)+k]->setFont(largefont);
          //field[(h*3*9)+(i*3)+(j*9)+k]->setTextColor(FXRGB(0,0,128));
          }
        }
      }
    }
  mainmatrix->setBackColor(FXRGB(0,0,0));
  }

SudokuWindow::~SudokuWindow(){
    delete filemenu;
    delete editmenu;
    delete helpmenu;
    delete largefont;
    delete smallfont;
    delete icon_exit;
    delete icon_undo;
    delete icon_redo;
    delete icon_logo;
    delete icon_logo_about;
    delete icon_logo_small;
    }

void SudokuWindow::create(){
    FXMainWindow::create();
    show(PLACEMENT_SCREEN);
    }


void SudokuWindow::init(){
  undolist.clear();
  for (int i=0;i<81;i++){
      if (sudoku_current.get(i)!=0) {
          field[i]->setEditable(false);
          field[i]->setText(FXStringVal(sudoku_current.get(i)));
          pv[i]->setText(FXString::null);
          }
      else {
        field[i]->setText(FXString::null);
        field[i]->setEditable(true);
        pv[i]->setText(FXString::null);
        }
      }
  }

bool SudokuWindow::saveGame(const FXString & filename){
  FXint version=20060211;  
  FXFileStream stream;
  if (stream.open(filename,FXStreamSave)) {
    stream << version;    
    sudoku_current.save(stream);        
    for (int i=0;i<81;i++){
      stream << pv[i]->getText();
      }
    stream.close();
    return true;
    }
  return false;
  }

bool SudokuWindow::loadGame(const FXString & filename){
  FXString text;
  FXint version;
  FXFileStream stream;
  if (stream.open(filename,FXStreamLoad)) {
    stream >> version;
    if (version==20060211){
      sudoku_puzzle.load(stream);      
      sudoku_current=sudoku_puzzle;
      init();
      for (int i=0;i<81;i++){
        stream >> text;
        pv[i]->setText(text);
        }  
      }
    else {
      FXMessageBox::error(this,MBOX_OK,"Unknown File","Error Reading File.");
      stream.close();
      return false;
      }
    stream.close();
    return true;
    }
  return false;
  }

void SudokuWindow::removePossibleValues(int i) {
  FXString text;
  int pos = 0;
  int row = i / 9;
  int col = i % 9;
  int sr = row / 3;
  int sc = col / 3;
  FXchar value = field[i]->getText()[0];

  /// Clear PV for i
  if (!pv[i]->getText().empty()){
    if (!undolist.busy()) undolist.add(new AnnotationMove(pv[i],pv[i]->getText()));
    pv[i]->setText(FXString::null);
    }

  /// Row
  for (int i=(row*9);i<(row*9)+9;i++){
    text = pv[i]->getText();
    while((pos=text.find(value))!=-1){
      text.erase(pos);
      }
    if (text!=pv[i]->getText()){
      if (!undolist.busy()) undolist.add(new AnnotationMove(pv[i],pv[i]->getText()));
      pv[i]->setText(text);
      }
    }

  /// Column
  for (int i=0;i<9;i++){
    text = pv[(i*9)+col]->getText();
    while((pos=text.find(value))!=-1){
      text.erase(pos);
      }
    if (text!=pv[(i*9)+col]->getText()){
      if (!undolist.busy()) undolist.add(new AnnotationMove(pv[(i*9)+col],pv[(i*9)+col]->getText()));
      pv[(i*9)+col]->setText(text);
      }
    }

  /// Section
  for (int i=0;i<3;i++){
    for (int j=0;j<3;j++){
      text = pv[(sr*3*9)+(sc*3)+(i*9)+j]->getText();
      while((pos=text.find(value))!=-1){
        text.erase(pos);
        }
      if (text!=pv[(sr*3*9)+(sc*3)+(i*9)+j]->getText()){
        if (!undolist.busy()) undolist.add(new AnnotationMove(pv[(sr*3*9)+(sc*3)+(i*9)+j],pv[(sr*3*9)+(sc*3)+(i*9)+j]->getText()));
        pv[(sr*3*9)+(sc*3)+(i*9)+j]->setText(text);
        }
      }
    }
  }



long SudokuWindow::onCmdField(FXObject*sender,FXSelector sel,void* ptr){
  if (FXString((const FXchar*)ptr).empty()) {

    if (FXSELTYPE(sel)==SEL_COMMAND && sudoku_current.get(FXSELID(sel)-ID_FIELD_0)==0)
      return 1;

    if (!undolist.busy())
      undolist.add(new SudokuMove(dynamic_cast<FXTextField*>(sender), sudoku_current.get(FXSELID(sel)-ID_FIELD_0)));

    sudoku_current.set(FXSELID(sel)-ID_FIELD_0,0);

//    fxmessage("Row: %d /  Col: %d =  - \n",(FXSELID(sel)-ID_FIELD_0)/9,(FXSELID(sel)-ID_FIELD_0)%9);
    }
  else {
    FXint value;
    sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_GETINTVALUE),&value);

    if (FXSELTYPE(sel)==SEL_COMMAND && sudoku_current.get(FXSELID(sel)-ID_FIELD_0)==value)
      return 1;

    if (!undolist.busy())
      undolist.begin(new FXCommandGroup);

    if (!undolist.busy())
      undolist.add(new SudokuMove(dynamic_cast<FXTextField*>(sender),sudoku_current.get(FXSELID(sel)-ID_FIELD_0)));

    sudoku_current.set(FXSELID(sel)-ID_FIELD_0,value);

    if (!undolist.busy()) {
      removePossibleValues(FXSELID(sel)-ID_FIELD_0);
      undolist.end();
      }

    if (sudoku_current.solved()){
      FXMessageBox::information(this,MBOX_OK,"Puzzle Solved!","Congratulations, you've solved the puzzle.");
      }


//    fxmessage("Row: %d /  Col: %d =  %d \n",(FXSELID(sel)-ID_FIELD_0)/9,(FXSELID(sel)-ID_FIELD_0)%9,value);
    }
  return 1;
  }

long SudokuWindow::onVerifyField(FXObject*sender,FXSelector sel,void* ptr){
    if (FXString((const FXchar*)ptr).empty()) {
        //fxmessage("Empty verify\n");
        sudoku_current.set(FXSELID(sel)-ID_FIELD_0,0);
        return 0;
        }
    FXint value = FXIntVal((FXchar*)ptr);
    if (value>0 && sudoku_current.is_valid(FXSELID(sel)-ID_FIELD_0,value))
      return 0;

    getApp()->beep();
    return 1;
    }

long SudokuWindow::onCmdAnnotation(FXObject*sender,FXSelector sel,void* ptr){
  if (!undolist.busy()) {
    FXString text;
    sender->handle(this,FXSEL(SEL_COMMAND,ID_GETSTRINGVALUE),&text);
    //fxmessage("Old: %s - New: %s\n",text.text(),(char*)ptr);
    undolist.add(new AnnotationMove(dynamic_cast<FXTextField*>(sender),text));
    }
  return 0; // SEL_VERIFY: no objection.
  }


long SudokuWindow::onCmdSolve(FXObject*,FXSelector,void*){
  if (FXMessageBox::question(this,MBOX_YES_NO,"Solve this Puzzle?","Are you sure you want to solve this puzzle?\nIt really takes all the fun out of the game.")==MBOX_CLICKED_YES){
    SudokuSolver solver;
    if (solver.solve(sudoku_current)){
      init();
      }
    else {
      FXMessageBox::error(this,MBOX_OK,"No Solution Found","Unable to find any solution for this puzzle");
      }
    }
  return 1;
  }

long SudokuWindow::onUpdSolve(FXObject*sender,FXSelector,void*){
  if (sudoku_current.solved())
    sender->handle(this,FXSEL(SEL_COMMAND,ID_DISABLE),NULL);
  else
    sender->handle(this,FXSEL(SEL_COMMAND,ID_ENABLE),NULL);
  return 1;
  }

long SudokuWindow::onUpdSave(FXObject*sender,FXSelector,void*){
  if (sudoku_current.empty())
    sender->handle(this,FXSEL(SEL_COMMAND,ID_DISABLE),NULL);
  else
    sender->handle(this,FXSEL(SEL_COMMAND,ID_ENABLE),NULL);
  return 1;
  }


long SudokuWindow::onCmdClear(FXObject*,FXSelector,void*){
  sudoku_current.clear();
  sudoku_puzzle.clear();
  init();
  return 1;
  }


long SudokuWindow::onCmdReset(FXObject*,FXSelector,void*){
  sudoku_current=sudoku_puzzle;  
  init();
  return 1;
  }



long SudokuWindow::onCmdLoad(FXObject*,FXSelector,void*){
  FXString filename = FXFileDialog::getOpenFilename(this,"Load Save Game",FXSystem::getHomeDirectory()+PATHSEPSTRING,"Sudoku Save Game (*.sdk)");
  if (!filename.empty()){
    loadGame(filename);
    }
  return 1;
  }

long SudokuWindow::onCmdSave(FXObject*,FXSelector,void*){
  FXString filename = FXFileDialog::getSaveFilename(this,"Save Game",FXSystem::getHomeDirectory()+PATHSEPSTRING,"Sudoku Save Game (*.sdk)");
  if (!filename.empty()){
    if (FXPath::extension(filename).empty())
      filename+=".sdk";
    saveGame(filename);
    }
  return 1;
  }


long SudokuWindow::onCmdAbout(FXObject*,FXSelector,void*){
  FXAboutDialog dialog(this,"Sudoku",APPLICATION_VERSION_STRING,"Sander Jansen","2006",icon_logo_about);
  dialog.execute();
  return 1;
  }


